//
//  coreJIT.cpp
//  guileJIT
//
//  Created by Andy Kitchen on 18/11/07.
//  Copyright (c) 2007 Andy Kitchen. All rights reserved.
//

#include "coreJIT.h"

FunctionBuilder::FunctionBuilder(Module *lModule): mModule(lModule) {
    std::vector<const Type*> ArgsToDouble(1, PointerType::get(Type::Int8Ty));
    FunctionType *FT1 = FunctionType::get(Type::DoubleTy, 
                                          ArgsToDouble, false);
    mUnpackValue = new Function(FT1, Function::ExternalLinkage, 
                                "scm_to_double", mModule);
    
    std::vector<const Type*> ArgsFromDouble(1, Type::DoubleTy);
    FunctionType *FT2 = FunctionType::get(PointerType::get(Type::Int8Ty), 
                                          ArgsFromDouble, false);
    mPackValue = new Function(FT2, Function::ExternalLinkage, 
                              "scm_from_double", mModule);
    
    mModuleProvider = new ExistingModuleProvider(mModule);
    mExecutionEngine = ExecutionEngine::create(mModuleProvider, false);
    
    mPassManager = new FunctionPassManager(mModuleProvider);
    mPassManager->add(createPromoteMemoryToRegisterPass());
}

void
FunctionBuilder::ImportFunction(const char* name, int num_args) {
    std::vector<const Type*> FuncArgs(num_args, PointerType::get(Type::Int8Ty));
    FunctionType *FT = FunctionType::get(PointerType::get(Type::Int8Ty), 
                                         FuncArgs, false);
    
    new Function(FT, Function::ExternalLinkage, 
                 name, mModule);
}

Function *
FunctionBuilder::CreateFunction(const char *name, SCM arg_ls, SCM body) {
    int arg_len = scm_to_int(scm_length(arg_ls));
    std::vector<const Type*> ArgsList(arg_len, PointerType::get(Type::Int8Ty));
    FunctionType *FTLocal = FunctionType::get(PointerType::get(Type::Int8Ty), 
                                          ArgsList, false);    
    
    Function *Subroutine =
    cast<Function>(mModule->getOrInsertFunction(name, FTLocal));
    
    BasicBlock *BB = new BasicBlock("EntryBlock", Subroutine);
    mBuilder.SetInsertPoint(BB);

    Function::arg_iterator AI = Subroutine->arg_begin();
    int i = 0;
    
    while(!(scm_is_null(arg_ls))) {
        SCM scm_str = scm_symbol_to_string(SCM_CAR(arg_ls));
        char * arg_name = scm_to_locale_string(scm_str);
        AI->setName(arg_name);
        
        string s_name(arg_name); s_name += "_store";
        AllocaInst *AllocSlot = 
        mBuilder.CreateAlloca(Type::DoubleTy, 0, s_name.c_str());
        mBindingTable[arg_name] = AllocSlot;

        string up_name(arg_name); up_name += "_unpacked";
        Value *UPValue = mBuilder.CreateCall(mUnpackValue, AI, up_name.c_str());
        mBuilder.CreateStore(UPValue, AllocSlot); 
        
        free(arg_name);
        i++, AI++;
        arg_ls = SCM_CDR(arg_ls);        
    }                                                   
        

    // Build The Body
    Value *EvalValue;
    try {
        EvalValue = CreateBody(body);
    } catch(std::exception e) {
        Subroutine->eraseFromParent();
        throw e;
    }
    
    CallInst *ReturnValue = mBuilder.CreateCall(mPackValue, 
                                                EvalValue, 
                                                "ReturnValue");
    mBuilder.CreateRet(ReturnValue);

    return Subroutine;
}

Value *
FunctionBuilder::CreateBody(SCM body) {
    if(scm_is_pair(body)) {
        if(scm_is_eq(SCM_CAR(body), scm_from_locale_symbol("+"))) {
            return mBuilder.CreateAdd(CreateBody(scm_cadr(body)), 
                                      CreateBody(scm_caddr(body)), "add_tmp");
        } else
        
        if(scm_is_eq(SCM_CAR(body), scm_from_locale_symbol("-"))) {
            return mBuilder.CreateSub(CreateBody(scm_cadr(body)), 
                                      CreateBody(scm_caddr(body)), "sub_tmp");
        } else

        if(scm_is_eq(SCM_CAR(body), scm_from_locale_symbol("*"))) {
            return mBuilder.CreateMul(CreateBody(scm_cadr(body)), 
                                      CreateBody(scm_caddr(body)), "mul_tmp");
        } else

        if(scm_is_eq(SCM_CAR(body), scm_from_locale_symbol("/"))) {
            return mBuilder.CreateFDiv(CreateBody(scm_cadr(body)), 
                                      CreateBody(scm_caddr(body)), "div_tmp");
        }
        
        else {
            SCM scm_str = scm_symbol_to_string(SCM_CAR(body));
            char * func_name = scm_to_locale_string(scm_str);
            
            Function *CallFunc = mModule->getFunction(func_name);
            
            if(CallFunc == NULL) {
                throw std::exception();
            }
            
            SCM scm_ptr = SCM_CDR(body);
            std::vector<Value*> ArgsV;
            while (!(scm_is_null(scm_ptr))) {
                Value *ArgValue = CreateBody(SCM_CAR(scm_ptr));
                Value *PackedArg = mBuilder.CreateCall(mPackValue, 
                                                       ArgValue, "pack_tmp");
                ArgsV.push_back(PackedArg);
                scm_ptr = SCM_CDR(scm_ptr);
            }
            
            Value *ReturnValue = mBuilder.CreateCall(CallFunc, 
                                                     ArgsV.begin(), 
                                                     ArgsV.end(), "return_tmp");            
            
            return mBuilder.CreateCall(mUnpackValue, 
                                       ReturnValue, "unpack_tmp");
            }
    } else {
        if (scm_is_number(body)) {
            return ConstantFP::get(Type::DoubleTy, scm_to_double(body));
        } else
        
        if (scm_is_symbol(body)) {
            SCM scm_str = scm_symbol_to_string(body);
            char * var_name = scm_to_locale_string(scm_str);
            return mBuilder.CreateLoad(mBindingTable[var_name]);
        }
    }
    
    return ConstantFP::get(Type::DoubleTy, 0.0);
}

jit_func_t
FunctionBuilder::CompileJIT(const char* name, SCM args, SCM body) {
    Function *Func = CreateFunction(name, args, body);
    mPassManager->run(*Func);    
        
    if (verifyModule(*mModule)) {
        std::cerr << "error: module verify failed" << std::endl;
        throw std::exception();
    }    
    
    Func->dump();

    return (jit_func_t)mExecutionEngine->getPointerToFunction(Func);
}